import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Sequence, Dict, Union, ClassVar
import numpy as np
from ...utils.periodic_table import PERIODIC_TABLE


class ANIAEV(nn.Module):
    """Computes the Atomic Environment Vector (AEV) for a given molecular system using the ANI model.

    FID : ANI_AEV

    ### Reference
    J. S. Smith, O. Isayev and A. E. Roitberg, ANI-1: an extensible neural network potential with DFT accuracy at force field computational cost, Chem. Sci., 2017, 8, 3192
    """

    _graphs_properties: Dict
    species_order: Union[str, Sequence[str]]
    """ The chemical species which are considered by the model."""
    graph_angle_key: str
    """ The key in the input dictionary that corresponds to the angular graph."""
    radial_eta: float = 16.0
    """ Controls the width of the gaussian sensitivity functions in radial AEV."""
    angular_eta: float = 8.0
    """ Controls the width of the gaussian sensitivity functions in angular AEV."""
    radial_dist_divisions: int = 16
    """ Number of basis function to encode distance in radial AEV."""
    angular_dist_divisions: int = 4
    """ Number of basis function to encode distance in angular AEV."""
    zeta: float = 32.0
    """ The power parameter in angle embedding."""
    angle_sections: int = 4
    """ The number of angle sections."""
    radial_start: float = 0.8
    """ The starting distance in radial AEV."""
    angular_start: float = 0.8
    """ The starting distance in angular AEV."""
    embedding_key: str = "embedding"
    """ The key to use for the output embedding in the returned dictionary."""
    graph_key: str = "graph"
    """ The key in the input dictionary that corresponds to the radial graph."""

    FID: ClassVar[str] = "ANI_AEV"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        rev_idx = {s: k for k, s in enumerate(PERIODIC_TABLE)}
        maxidx = max(rev_idx.values())

        # convert species to internal indices
        conv_tensor = [0] * (maxidx + 2)
        if isinstance(self.species_order, str):
            species_order = [el.strip() for el in self.species_order.split(",")]
        else:
            species_order = [el for el in self.species_order]
        for i, s in enumerate(species_order):
            conv_tensor[rev_idx[s]] = i
        indices = jnp.asarray(conv_tensor, dtype=jnp.int32)[species]
        num_species = len(species_order)
        num_species_pair = (num_species * (num_species + 1)) // 2

        # Radial graph
        graph = inputs[self.graph_key]
        distances = graph["distances"]
        switch = graph["switch"]
        edge_src = graph["edge_src"]
        edge_dst = graph["edge_dst"]

        # Radial AEV
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        shiftR = jnp.asarray(
            - np.linspace(self.radial_start, cutoff, self.radial_dist_divisions + 1)[
                None, :-1
            ],
            dtype=distances.dtype,
        )
        x2 = self.radial_eta * (distances[:, None] + shiftR) ** 2
        radial_terms = jnp.exp(-x2) * (0.25*switch)[:, None]
        # aggregate radial AEV
        radial_index = edge_src * num_species + indices[edge_dst]

        radial_aev = jax.ops.segment_sum(
            radial_terms, radial_index, num_species * species.shape[0]
        ).reshape(species.shape[0], num_species * radial_terms.shape[-1])

        # Angular graph
        # eta2 = self.angular_eta ** 0.5
        graph = inputs[self.graph_angle_key]
        angles = graph["angles"]
        distances = (0.5*self.angular_eta**0.5)*graph["distances"]
        central_atom = graph["central_atom"]
        angle_src, angle_dst = graph["angle_src"], graph["angle_dst"]
        d12 = (distances[angle_src] + distances[angle_dst])[:, None]

        # Angular AEV parameters
        angular_cutoff = self._graphs_properties[self.graph_angle_key]["cutoff"]
        angle_start = np.pi / (2 * self.angle_sections)
        shiftZ = jnp.asarray(
            -(np.linspace(0, np.pi, self.angle_sections + 1) + angle_start)[None, :-1],
            dtype=distances.dtype,
        )
        shiftA = jnp.asarray(
            (-self.angular_eta**0.5)*np.linspace(
                self.angular_start, angular_cutoff, self.angular_dist_divisions + 1
            )[None, :-1],
            dtype=distances.dtype,
        )

        # Angular AEV
        switch = graph["switch"] *(2.0 *(0.5**self.zeta))**0.5
        factor1 = switch[angle_src] * switch[angle_dst]
        factor1 = (
            factor1[:, None] * (1 + jnp.cos(angles[:, None] + shiftZ)) ** self.zeta
        )
        factor2 = jnp.exp(-(d12 + shiftA) ** 2)
        angular_terms = (factor1[:, None, :] * factor2[:, :, None]).reshape(
            -1, self.angle_sections * self.angular_dist_divisions
        )

        # aggregate angular AEV
        index_dest = indices[graph["edge_dst"]]
        species1, species2 = np.triu_indices(num_species, 0)
        pair_index = np.arange(species1.shape[0], dtype=np.int32)
        triu_index = np.zeros((num_species, num_species), dtype=np.int32)
        triu_index[species1, species2] = pair_index
        triu_index[species2, species1] = pair_index
        triu_index = jnp.asarray(triu_index, dtype=jnp.int32)
        angular_index = (
            central_atom * num_species_pair
            + triu_index[index_dest[angle_src], index_dest[angle_dst]]
        )

        angular_aev = jax.ops.segment_sum(
            angular_terms, angular_index, num_species_pair * species.shape[0]
        ).reshape(species.shape[0], num_species_pair * angular_terms.shape[-1])

        embedding = jnp.concatenate((radial_aev, angular_aev), axis=-1)
        if self.embedding_key is None:
            return embedding
        return {**inputs, self.embedding_key: embedding}



#!/usr/bin/env python3
"""Electric field model for FENNOL.

Created by C. Cattin 2024
"""

import flax.linen as nn
import jax
import jax.numpy as jnp
from typing import ClassVar

from fennol.utils import AtomicUnits as Au


class ElectricField(nn.Module):
    """Electric field from distributed point charges with short-range damping.

    FID: ELECTRIC_FIELD

    The short-range damping is defined as in AMOEBA+

    """

    damping_param: float = 0.7
    """Damping parameter for the electric field."""
    charges_key: str = 'charges'
    """Key of the charges in the input."""
    graph_key: str = 'graph'
    """Key of the graph in the input."""
    polarizability_key: str = 'polarizability'
    """Key of the polarizability in the input."""
    trainable: bool = False

    FID: ClassVar[str] = 'ELECTRIC_FIELD'

    @nn.compact
    def __call__(self, inputs):
        species = inputs['species']

        # Graph information
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph['edge_src'], graph['edge_dst']

        # Distance and vector between each pair of atoms in atomic units
        distances = graph['distances']
        rij = distances / Au.BOHR
        vec_ij = graph['vec'] / Au.BOHR
        polarizability = (
            inputs[self.polarizability_key] / Au.BOHR**3
        )
        pol_src = polarizability[edge_src]
        pol_dst = polarizability[edge_dst]
        alpha_ij = pol_dst * pol_src
        # Effective distance
        uij = rij / alpha_ij ** (1 / 6)

        # Charges and polarizability
        charges = inputs[self.charges_key]
        rij = rij[:, None]
        q_ij = charges[edge_dst, None]

        if self.trainable:
            damping_param = jnp.abs(self.param('damping_param', lambda key: jnp.array(self.damping_param)))
        else:
            damping_param = self.damping_param
        # Damping term
        damping_field = 1 - jnp.exp(
            -damping_param * uij**1.5
        )[:, None]

        # Electric field
        eij = -q_ij * (vec_ij / rij**3) * damping_field
        electric_field = jax.ops.segment_sum(
            eij, edge_src, species.shape[0]
        ).flatten()

        output = {
            'electric_field': electric_field
        }

        return {**inputs, **output}


if __name__ == "__main__":
    pass



""" 
# Available Modules
Available modules are listed in the `fennol.models.modules.MODULES` dictionary.
They can also be found by searching for 'FID' in the documentation's search bar.
"""

from .fennix import FENNIX
from .modules import register_fennix_module, get_modules_documentation, available_fennix_modules,MODULES



import numpy as np
import math

def get_reciprocal_space_parameters(reciprocal_cells, cutoff, kmax=30, kthr=1.0e-6):
    # find optimal ewald parameters (preprocessing)
    eps = 1.0e-8
    ratio = eps + 1
    x = 0.5
    i = 0
    # approximate value
    while ratio > eps:
        x *= 2
        ratio = math.erfc(x * cutoff) / cutoff
    # refine with binary search
    k = i + 60
    xlo = 0.0
    xhi = x
    for i in range(1, k + 1):
        x = (xlo + xhi) / 2.0
        ratio = math.erfc(x * cutoff) / cutoff
        if ratio > eps:
            xlo = x
        else:
            xhi = x
    bewald = x

    # set k points
    kxs = np.arange(kmax + 1)
    kxs = np.concatenate((kxs, -kxs[1:]))
    k = np.array(np.meshgrid(kxs, kxs, kxs)).reshape(3, -1).T[1:]
    # set exp factor
    m2s = []
    expfacs = []
    ks = []
    nks = []
    for i, A in enumerate(range(reciprocal_cells.shape[0])):
        A = reciprocal_cells[i].T
        m2 = np.sum(
            (
                k[:, 0, None] * A[None, 0, :]
                + k[:, 1, None] * A[None, 1, :]
                + k[:, 2, None] * A[None, 2, :]
            )
            ** 2,
            axis=-1,
        )
        a2 = (np.pi / bewald) ** 2
        expfac = np.exp(-a2 * m2) / m2
        isort = np.argsort(expfac)[::-1]
        expfac = expfac[isort]
        m2 = m2[isort]
        ki = k[isort]
        sel = (expfac > kthr).nonzero()[0]
        nks.append(len(sel))
        m2s.append(m2)
        expfacs.append(expfac)
        ks.append(ki)

    ks = np.stack(ks, axis=0)
    m2s = np.stack(m2s, axis=0)
    expfacs = np.stack(expfacs, axis=0)
    nks = np.array(nks, dtype=np.int64)
    nk = np.max(nks)
    return ks[:, :nk, :], expfacs[:, :nk], m2s[:, :nk], bewald



from typing import Callable, Optional, Dict, List, Tuple, Union, Any, NamedTuple
import optax
import jax
import jax.numpy as jnp
import numpy as np
import operator
from flax import traverse_util
import json
import re

from optax._src import base
from optax._src import wrappers



class AddWeightDiffState(NamedTuple):
    ref_weights: Any

def add_weights_difference(
    weight_decay: Union[float, jax.Array] = 0.0,
    mask: Optional[Union[Any, Callable[[base.Params], Any]]] = None
) -> base.GradientTransformation:
  """weight decay toward initial weights."""
  def init_fn(params):
    return AddWeightDiffState(ref_weights=params)

  def update_fn(updates, state, params):
    if params is None:
      raise ValueError(base.NO_PARAMS_MSG)
    updates = jax.tree_util.tree_map(
        lambda g, p, pref: g + weight_decay * (p-pref), updates, params,state.ref_weights)
    return updates, state

  # If mask is not `None`, apply mask to the gradient transformation.
  # E.g. it is common to skip weight decay on bias units and batch stats.
  if mask is not None:
    return wrappers.masked(
        base.GradientTransformation(init_fn, update_fn), mask)
  return base.GradientTransformation(init_fn, update_fn)



def add_grokfast(
    alpha: float = 0.9,
    l: float = 1.,
)-> base.GradientTransformation:
    """Grokfast: amplify slow gradients by exponential moving average."""

    ema: base.GradientTransformation = optax.ema(decay=alpha,debias=False)
    def init_fn(params):
        return ema.init(params)
    
    def update_fn(updates, state, params=None):
        dupdates, state = ema.update(updates, state, params)
        # updates = updates + l*dupdates
        updates = jax.tree_util.tree_map(
            lambda g,d: g+l*d, updates, dupdates)
        return updates, state
    
    return base.GradientTransformation(init_fn, update_fn)


class PROFITState(NamedTuple):
    ref_weights: Any
    istep: int
    main_opt_state: Any
    internal_opt_state: Any

def profit(
    learning_rate: base.ScalarOrSchedule,
    nsteps_ref: int = 1,
    main_opt: str = 'adam',
    main_opt_params: Dict[str, Any] = {},
    internal_opt: str = 'sgd',
    internal_opt_params: Dict[str, Any] = {},
    **kwargs
)-> base.GradientTransformation:
    """PROFIT optimizer for fine-tuning https://arxiv.org/pdf/2412.01930"""

    main_opt_params = {'learning_rate': learning_rate,**main_opt_params}
    main_opt = eval(
        main_opt,
        {"__builtins__": None},
        {**optax.__dict__},
    )(**main_opt_params)

    internal_opt_params = {'learning_rate': .1,**internal_opt_params}
    internal_opt = eval(
        internal_opt,
        {"__builtins__": None},
        {**optax.__dict__},
    )(**internal_opt_params)


    def init_fn(params):
        return PROFITState(
            ref_weights=params,
            istep=0,
            main_opt_state=main_opt.init(params),
            internal_opt_state=internal_opt.init(params),
        )
    

    def update_main(gradients,main_opt_state,internal_opt_state,params,params_ref):
        delta = jax.tree_util.tree_map(lambda p, pref: p-pref, params, params_ref)
        dot = jax.tree.reduce(
           operator.add,
           jax.tree_util.tree_map(lambda g,d: (g*d).sum(), gradients, delta)
        )
        delta2 = jax.tree.reduce(
           operator.add,
           jax.tree_util.tree_map(lambda d: (d**2).sum(), delta)
        )
        proj = dot/(delta2+1.e-6)

        gradients = jax.lax.cond(dot>=0,
            lambda g,d: g,
            lambda g,d: jax.tree_util.tree_map(lambda x: proj*x, d),
            gradients,delta
        )
        updates,main_opt_state = main_opt.update(gradients,main_opt_state,params)
        updates = jax.tree_util.tree_map(lambda g,d: g-d, updates, delta)
        return updates,main_opt_state,internal_opt_state
    
    def update_internal(gradients,main_opt_state,internal_opt_state,params,params_ref):
        updates,internal_opt_state = internal_opt.update(gradients,internal_opt_state,params)
        return updates,main_opt_state,internal_opt_state
    
    def update_fn(gradients, state, params):
        istep = state.istep % (nsteps_ref+1)
        # jax.debug.print("{i} {j}",i=istep,j=state.istep)

        params_ref = jax.lax.cond(istep==0, lambda a,b: a, lambda a,b: b, params, state.ref_weights)

        updates,main_opt_state,internal_opt_state = jax.lax.cond(
            istep==nsteps_ref,
            update_main,
            update_internal,
            gradients,state.main_opt_state,state.internal_opt_state,params,params_ref
        )

        new_state = PROFITState(
            ref_weights=params_ref,
            istep=state.istep+1,
            main_opt_state=main_opt_state,
            internal_opt_state=internal_opt_state
        )
        return updates, new_state

    return base.GradientTransformation(init_fn, update_fn)

def get_optimizer(
    training_parameters: Dict[str, any], variables: Dict, initial_lr: float
) -> optax.GradientTransformation:
    """
    Returns an optax.GradientTransformation object that can be used to optimize the model parameters.

    Args:
    - training_parameters: A dictionary containing the training parameters.
    - variables: A  pytree containing the model parameters.
    - initial_lr: The initial learning rate.

    Returns:
    - An optax.GradientTransformation object that can be used to optimize the model parameters.
    """

    default_status = str(training_parameters.get("default_status", "trainable")).lower()
    assert default_status in [
        "trainable",
        "frozen",
    ], f"Default status must be 'trainable' or 'frozen', got {default_status}"

    # find frozen and trainable parameters
    frozen = training_parameters.get("frozen", [])
    trainable = training_parameters.get("trainable", [])

    def training_status(full_path, v):
        full_path = "/".join(full_path[1:]).lower()
        # full_path = "/".join(full_path[1:]).lower()
        status = (default_status, "")
        for path in frozen:
            # if full_path.startswith(path.lower()) and len(path) > len(status[1]):
            if re.match(path.lower(), full_path):
                status = ("frozen", path)
        for path in trainable:
            # if full_path.startswith(path.lower()) and len(path) > len(status[1]):
            if re.match(path.lower(), full_path):
                status = ("trainable", path)
        return status[0]

    params_partition = traverse_util.path_aware_map(training_status, variables)
    if len(frozen) > 0 or len(trainable) > 0:
        print("params partition:")
        print(json.dumps(params_partition, indent=2, sort_keys=False))

    ## Gradient preprocessing
    grad_processing = []

    # zero nans
    zero_nans = training_parameters.get("zero_nans", False)
    if zero_nans:
        grad_processing.append(optax.zero_nans())

    use_grokfast = training_parameters.get("use_grokfast", False)
    if use_grokfast:
        print("Using Grokfast")
        alpha_grokfast = training_parameters.get("alpha_grokfast", 0.9)
        l_grokfast = training_parameters.get("l_grokfast", 1.0)
        grad_processing.append(add_grokfast(alpha=alpha_grokfast, l=l_grokfast))


    # gradient clipping
    clip_threshold = training_parameters.get("gradient_clipping", -1.0)
    if clip_threshold > 0.0:
        print("Adaptive gradient clipping threshold:", clip_threshold)
        grad_processing.append(optax.adaptive_grad_clip(clip_threshold))

    # OPTIMIZER
    optimizer_name = training_parameters.get("optimizer", "adabelief")
    optimizer = eval(
        optimizer_name,
        {"__builtins__": None},
        {**optax.__dict__,"profit":profit},
    )
    print("Optimizer:", optimizer_name)
    optimizer_configuration = training_parameters.get("optimizer_config", {})
    optimizer_configuration["learning_rate"] = 1.0
    grad_processing.append(optimizer(**optimizer_configuration))

    # weight decay
    weight_decay = training_parameters.get("weight_decay", 0.0)
    assert weight_decay >= 0.0, "Weight decay must be positive"
    decay_targets = training_parameters.get("decay_targets", [""])

    def decay_status(full_path, v):
        full_path = "/".join(full_path).lower()
        status = False
        # print(full_path,re.match(r'^params\/', full_path))
        for path in decay_targets:
            if re.match(r"^params/" + path.lower(), full_path):
                status = True
            # if full_path.startswith("params/" + path.lower()):
            #     status = True
        return status

    decay_mask = traverse_util.path_aware_map(decay_status, variables)
    if weight_decay > 0.0:
        print("weight decay:", weight_decay)
        print(json.dumps(decay_mask, indent=2, sort_keys=False))
        grad_processing.append(
            optax.add_decayed_weights(weight_decay=-weight_decay, mask=decay_mask)
        )
    
    regularize_init_weight = training_parameters.get("regularize_init_weights", 0.)
    if regularize_init_weight > 0.0:
        print("Regularizing toward initial weights with L2 norm:", 
        regularize_init_weight)
        if weight_decay <=0.:
            print(json.dumps(decay_mask, indent=2, sort_keys=False))

        grad_processing.append(add_weights_difference(weight_decay=-regularize_init_weight, mask=decay_mask))

    if zero_nans:
        grad_processing.append(optax.zero_nans())

    # learning rate
    grad_processing.append(optax.inject_hyperparams(optax.scale)(step_size=initial_lr))

    ## define optimizer chain
    optimizer_ = optax.chain(
        *grad_processing,
    )
    partition_optimizer = {"trainable": optimizer_, "frozen": optax.set_to_zero()}
    return optax.multi_transform(partition_optimizer, params_partition)

def get_lr_schedule(max_epochs,nbatch_per_epoch,training_parameters):
    lr = training_parameters.get("lr", 1.0e-3)
    init_lr = training_parameters.get("init_lr", lr / 25)
    final_lr = training_parameters.get("final_lr", lr / 10000)

    #### LEARNING RATE SCHEDULER ####
    schedule_type = training_parameters.get("schedule_type", "cosine_onecycle").lower()
    schedule_type = training_parameters.get("scheduler", schedule_type).lower()
    schedule_metrics = training_parameters.get("schedule_metrics", "rmse_tot")

    adaptive_scheduler = False
    print("Schedule type:", schedule_type)
    if schedule_type == "cosine_onecycle":
        transition_epochs = training_parameters.get("onecycle_epochs", max_epochs)
        peak_epoch = training_parameters.get("peak_epoch", 0.3 * transition_epochs)
        schedule_ = optax.cosine_onecycle_schedule(
            peak_value=lr,
            div_factor=lr / init_lr,
            final_div_factor=init_lr / final_lr,
            transition_steps=transition_epochs * nbatch_per_epoch,
            pct_start=peak_epoch / transition_epochs,
        )
        sch_state = {"count": 0, "best": np.inf, "lr": init_lr}

        def schedule(state, rmse=None):
            new_state = {**state}
            lr = schedule_(state["count"])
            if rmse is None:
                new_state["count"] += 1
            new_state["lr"] = lr
            return lr, new_state
    
    elif schedule_type == "piecewise_interpolate":
        schedule_params = training_parameters.get("scheduler_parameters", {})
        schedule_ = optax.piecewise_interpolate_schedule(
            **{"init_value":lr,"interpolate_type":"linear",**schedule_params}
        )
        sch_state = {"count": 0, "best": np.inf, "lr": schedule_(0)}
        def schedule(state, rmse=None):
            new_state = {**state}
            lr = schedule_(state["count"])
            if rmse is None:
                new_state["count"] += 1
            new_state["lr"] = lr
            return lr, new_state

    elif schedule_type == "constant":
        sch_state = {"count": 0}

        def schedule(state, rmse=None):
            new_state = {**state}
            new_state["lr"] = lr
            if rmse is None:
                new_state["count"] += 1
            return lr, new_state

    elif schedule_type == "reduce_on_plateau":
        patience = training_parameters.get("patience", 10)
        factor = training_parameters.get("lr_factor", 0.5)
        patience_thr = training_parameters.get("patience_thr", 0.0)
        sch_state = {"count": 0, "best": np.inf, "lr": lr, "patience": patience}
        adaptive_scheduler = True

        def schedule(state, rmse=None):
            new_state = {**state}
            if rmse is None:
                new_state["count"] += 1
                return state["lr"], new_state
            if rmse <= state["best"] * (1.0 + patience_thr):
                if rmse < state["best"]:
                    new_state["best"] = rmse
                new_state["patience"] = 0
            else:
                new_state["patience"] += 1
                if new_state["patience"] >= patience:
                    new_state["lr"] = state["lr"] * factor
                    new_state["patience"] = 0
                    print("Reducing learning rate to", new_state["lr"])
            return new_state["lr"], new_state

    else:
        raise ValueError(f"Unknown schedule_type: {schedule_type}")

    stochastic_scheduler = training_parameters.get("stochastic_scheduler", False)
    if stochastic_scheduler:
        schedule_ = schedule
        rng_key, scheduler_key = jax.random.split(rng_key)
        sch_state["rng_key"] = scheduler_key
        sch_state["lr_max"] = lr
        sch_state["lr_min"] = final_lr

        def schedule(state, rmse=None):
            new_state = {**state, "lr": state["lr_max"]}
            if rmse is None:
                lr_max, new_state = schedule_(new_state, rmse=rmse)
                lr_min = new_state["lr_min"]
                new_state["rng_key"], subkey = jax.random.split(new_state["rng_key"])
                lr = lr_min + (lr_max - lr_min) * jax.random.uniform(subkey)
                new_state["lr"] = lr
                new_state["lr_max"] = lr_max

            return new_state["lr"], new_state
    
    return schedule, sch_state, schedule_metrics, adaptive_scheduler


import jax
import jax.numpy as jnp
from functools import partial
import math
import flax.linen as nn
import numpy as np
from typing import Union, Callable


class TrainableSiLU(nn.Module):
    @nn.compact
    def __call__(self, x):
        a = self.param("alpha", lambda k, s: jnp.ones(s), (1, x.shape[-1]))
        b = self.param("beta", lambda k, s: 1.702 * jnp.ones(s), (1, x.shape[-1]))
        shape = x.shape
        x = x.reshape(-1, shape[-1])
        return (a * x * jax.nn.sigmoid(b * x)).reshape(shape)

class DyT(nn.Module):
    alpha0: float = 0.5

    @nn.compact
    def __call__(self, x):
        shape = x.shape
        dim = shape[-1]
        a = self.param("scale", lambda k, s: jnp.ones(s,dtype=x.dtype), (1, dim))
        b = self.param("bias", lambda k, s: jnp.zeros(s,dtype=x.dtype), (1, dim))
        alpha = self.param("alpha", lambda k: jnp.asarray(self.alpha0,dtype=x.dtype))
        x = x.reshape(-1, dim)
        return (a*jnp.tanh(alpha*x) + b).reshape(shape)

class DynAct(nn.Module):
    activation : Union[str, Callable]
    alpha0: float = 1.
    channelwise: bool = False

    @nn.compact
    def __call__(self, x):
        shape = x.shape
        dim = shape[-1]
        a = self.param("scale", lambda k, s: jnp.ones(s,dtype=x.dtype), (1, dim))
        b = self.param("bias", lambda k, s: jnp.zeros(s,dtype=x.dtype), (1, dim))

        shape_alpha = (1,dim) if self.channelwise else (1,1)
        alpha = self.param("alpha", lambda k,s: self.alpha0*jnp.ones(s,dtype=x.dtype), shape_alpha)

        x = x.reshape(-1, dim)
        act = activation_from_str(self.activation)
        return (a*act(alpha*x) + b).reshape(shape)

class TrainableCELU(nn.Module):
    alpha: float = 0.1

    @nn.compact
    def __call__(self, x):
        a = self.alpha * (
            1
            + jax.nn.celu(
                self.param(
                    "alpha",
                    lambda k, s: jnp.zeros(s),
                    (
                        1,
                        x.shape[-1],
                    ),
                ),
                alpha=1.0,
            )
        )
        shape = x.shape
        x = x.reshape(-1, shape[-1])
        return jax.nn.celu(x, a).reshape(shape)


class TrainableLeakyCELU(nn.Module):
    alpha: float = 0.05
    beta: float = 1.0

    @nn.compact
    def __call__(self, x):
        a = self.alpha + self.param(
            "alpha",
            lambda k, s: jnp.zeros(s),
            (
                1,
                x.shape[-1],
            ),
        )
        b = self.beta * (
            1
            + jax.nn.celu(
                self.param(
                    "beta",
                    lambda k, s: jnp.zeros(s),
                    (
                        1,
                        x.shape[-1],
                    ),
                ),
                alpha=1.0,
            )
        )
        shape = x.shape
        x = x.reshape(-1, shape[-1])
        return leaky_celu(x, a, b).reshape(shape)


class FourierActivation(nn.Module):
    nmax: int = 4

    @nn.compact
    def __call__(self, x):

        out = jax.nn.swish(x)
        nfeatures = x.shape[-1]
        n = 2 * np.pi * np.arange(1, self.nmax + 1)
        shape = [1] * x.ndim + [self.nmax]
        n = n.reshape(shape)
        x = jnp.expand_dims(x, axis=-1)
        cx = jnp.cos(n * x)
        sx = jnp.sin(n * x)
        x = jnp.concatenate((cx, sx), axis=-1)

        shape = [1] * (out.ndim - 1) + [nfeatures]
        b = self.param("b", jax.nn.initializers.zeros, shape)

        w = self.param(
            "w", jax.nn.initializers.normal(stddev=0.1), (*shape, 2 * self.nmax)
        )

        return out + (x * w).sum(axis=-1) + b


class GausianBasisActivation(nn.Module):
    nbasis: int = 10
    xmin: float = -1.0
    xmax: float = 3.0

    @nn.compact
    def __call__(self, x):

        nfeatures = x.shape[-1]
        shape = [1] * (x.ndim - 1) + [nfeatures]
        b = self.param("b", jax.nn.initializers.zeros, shape)
        w0 = self.param("w0", jax.nn.initializers.ones, shape)
        out = jax.nn.swish(w0*x)

        shape+= [self.nbasis]
        x0 = self.param(
            "x0",
            lambda key: jnp.asarray(
                np.linspace(self.xmin, self.xmax, self.nbasis)[None, :]
                .repeat(nfeatures, axis=0)
                .reshape(shape),
                dtype=x.dtype,
            ),
        )

        sigma0 = np.abs(self.xmax - self.xmin)/self.nbasis
        alphas = (0.5**0.5)*self.param(
            "sigmas",
            lambda key: jnp.full(shape, 1./sigma0, dtype=x.dtype),
        )

        w = self.param(
            "w", jax.nn.initializers.normal(stddev=0.1), shape
        )

        ex = jnp.exp(-((x[...,None] - x0) * alphas)**2)

        return out + (ex * w).sum(axis=-1) + b

def safe_sqrt(x,eps=1.e-5):
    return jnp.sqrt(jnp.clip(x, min=eps))

@jax.jit
def aptx(x):
    return (1.0 + jax.nn.tanh(x)) * x


@jax.jit
def serf(x):
    return x * jax.scipy.special.erf(jax.nn.softplus(x))


@partial(jax.jit, static_argnums=(1,))
def leaky_celu(x, alpha=0.1, beta=1.0):
    return alpha * x + ((1.0 - alpha) / beta) * (
        jax.nn.softplus(beta * x) - math.log(2.0)
    )


@jax.jit
def tssr(x):
    ax = jnp.abs(x)
    mask = ax <= 1.0
    axx = jnp.where(mask, 1.0, ax)
    return jnp.where(mask, x, jnp.sign(x) * (2 * axx**0.5 - 1))


@jax.jit
def tssr2(x):
    ax = jnp.abs(x)
    mask = ax <= 1.0
    axx = jnp.where(mask, 1.0, ax)
    return jnp.sign(x) * jnp.where(mask, 1.25 * ax - 0.25 * ax**3, axx**0.5)


@jax.jit
def tssr3(x):
    ax = jnp.abs(x)
    mask = ax <= 1.0
    axx = jnp.where(mask, 1.0, ax)
    ax2 = ax * ax
    dax2 = ax - ax2
    poly = 2.1875 * dax2 + ax2 * (ax + 0.3125 * dax2)
    return jnp.sign(x) * jnp.where(mask, poly, axx**0.5)


@jax.jit
def pow(x, a):
    return x**a


@jax.jit
def ssp(x):
    return jnp.logaddexp(x + math.log(0.5), math.log(0.5))


@jax.jit
def smooth_floor(x, eps=0.99):
    return (
        x
        - 0.5
        - jnp.atan(
            -eps * jnp.sin((-2 * jnp.pi) * x) / (eps * jnp.cos((2 * jnp.pi) * x) - 1.0)
        )
        / jnp.pi
    )


@jax.jit
def smooth_round(x, eps=0.99):
    return (
        x
        - jnp.atan(
            -eps
            * jnp.sin(-2 * jnp.pi * (x - 0.5))
            / (eps * jnp.cos(2 * jnp.pi * (x - 0.5)) - 1.0)
        )
        / jnp.pi
    )


def chain(*activations):
    # @jax.jit
    def act(x):
        for a in activations:
            x = a(x)
        return x

    return act


def normalize_activation(
    phi: Callable[[float], float], return_scale=False
) -> Callable[[float], float]:
    r"""Normalize a function, :math:`\psi(x)=\phi(x)/c` where :math:`c` is the normalization constant such that

    .. math::

        \int_{-\infty}^{\infty} \psi(x)^2 \frac{e^{-x^2/2}}{\sqrt{2\pi}} dx = 1

    ! Adapted from e3nn_jax !
    """
    with jax.ensure_compile_time_eval():
        # k = jax.random.PRNGKey(0)
        # x = jax.random.normal(k, (1_000_000,))
        n = 1_000_001
        x = jnp.sqrt(2) * jax.scipy.special.erfinv(jnp.linspace(-1.0, 1.0, n + 2)[1:-1])
        c = jnp.mean(phi(x) ** 2) ** 0.5
        c = c.item()

        if jnp.allclose(c, 1.0):
            rho = phi
        else:

            def rho(x):
                return phi(x) / c

        if return_scale:
            return rho, 1.0 / c
        return rho


def activation_from_str(activation: Union[str, Callable, None]) -> Callable:
    if activation is None:
        return lambda x: x
    if callable(activation):
        return activation
    if not isinstance(activation, str):
        raise ValueError(f"Invalid activation {activation}")
    if activation.lower() in ["none", "linear", "identity"]:
        return lambda x: x
    try:
        return eval(
            activation,
            {"__builtins__": None},
            {
                **jax.nn.__dict__,
                **jax.numpy.__dict__,
                **jax.__dict__,
                "chain": chain,
                "pow": pow,
                "partial": partial,
                "leaky_celu": leaky_celu,
                "aptx": aptx,
                "tssr": tssr,
                "tssr2": tssr2,
                "tssr3": tssr3,
                "ssp": ssp,
                "smooth_floor": smooth_floor,
                "smooth_round": smooth_round,
                "TrainableSiLU": TrainableSiLU,
                "TrainableCELU": TrainableCELU,
                "FourierActivation": FourierActivation,
                "GaussianBasis": GausianBasisActivation,
                "normalize": normalize_activation,
                "DyT": DyT,
                "DynAct": DynAct,
            },
        )
    except Exception as e:
        raise ValueError(
            f"The following exception was raised while parsing the activation function {activation} : {e}"
        )
